CREATE OR REPLACE PACKAGE plpdf_xfdf is
--v2.0.0
/* Simple XFDF Composer and Parser*/

-- type
type t_xml_inputs is table of plx_sc.maxv2 index by plx_sc.normalv2;

-- variable
g_xml_doc clob;

-- constants
c_xml_version    constant plx_sc.maxv2 := '1.0';
c_xml_encoding   constant plx_sc.maxv2 := 'utf-8';
c_xml_standalone constant plx_sc.maxv2 := null;
c_xmlns          constant plx_sc.maxv2 := 'http://ns.adobe.com/xfdf/';
c_xml_space      constant plx_sc.maxv2 := 'preserve';
t_xmlns    constant plx_sc.normalv2 := 'xmlns';
t_xml      constant plx_sc.normalv2 := 'xml';
t_space    constant plx_sc.normalv2 := 'space';
t_xfdf     constant plx_sc.normalv2 := 'xfdf';
t_original constant plx_sc.normalv2 := 'original';
t_modified constant plx_sc.normalv2 := 'modified';
t_ids      constant plx_sc.normalv2 := 'ids';
t_fields   constant plx_sc.normalv2 := 'fields';
t_name     constant plx_sc.normalv2 := 'name';
t_field    constant plx_sc.normalv2 := 'field';
t_value    constant plx_sc.normalv2 := 'value';

/* Subprograms */

-- Start the document
procedure start_;

-- End the document
procedure end_;

-- Open xfdf tag
procedure startxfdf;

-- Close xfdf tag
procedure endxfdf;

-- Add ids tag with original and modified attributes
procedure ids(
  p_original_id plx_sc.maxv2, 
  p_modified_id plx_sc.maxv2
  );

-- Open fields tag
procedure startfields;

-- Close fields tag
procedure endfields;

-- Open field tag and add name attribute
procedure startfield(
  p_name plx_sc.maxv2
  );

-- Close field tag
procedure endfield;

-- Open value tag
procedure startvalue;

-- Close value tag
procedure endvalue;

-- Add value
procedure value(
    p_value plx_sc.maxv2
);

-- Get blob type from clob
function getblob(
  p_xml_doc clob
  ) return blob;

/*
-TEST CREATE
declare
v_nodes plpdf_xfdf.t_xml_inputs;
v_blob blob;
begin
v_nodes('CAR') := 'Ford';
v_nodes('MOTORBIKE') := 'Suzuki';
v_nodes('BIKE') := 'KTM';
v_nodes('BOAT') := 'Bayliner';
v_blob := plpdf_xfdf.create_xfdf(v_nodes, sys_guid());
end;
*/

-- Generate an XFDF document from PL/SQL table
function create_xfdf(
  p_xml_inputs t_xml_inputs,
  p_id varchar2 default null
  ) return blob;

/*
-TEST PARSE
declare
p_file blob;
v_clob clob;
v_varchar varchar2(32767);
v_start pls_integer := 1;
v_buffer pls_integer := 32767;
xml_ret plpdf_xfdf.t_xml_inputs;
v_i plx_sc.normalv2;
begin
    select blob_file into p_file from store_blob_tk where filename = 'test_input.xml';
    dbms_lob.createtemporary(v_clob, true);
  
    for i in 1..ceil(dbms_lob.getlength(p_file) / v_buffer)
        loop
            v_varchar := utl_raw.cast_to_varchar2(dbms_lob.substr(p_file, v_buffer, v_start));
            dbms_lob.writeappend(v_clob, length(v_varchar), v_varchar);
            v_start := v_start + v_buffer;
        end loop;
        
    xml_ret := plpdf_xfdf.parse_xfdf(v_clob);
    
    dbms_output.enable(1000000);
    v_i := xml_ret.first;
    while v_i is not null loop
        dbms_output.put_line('xml_ret(' || v_i || ') = ' || xml_ret(v_i));
        v_i := xml_ret.next(v_i);
    end loop;
end;
*/
-- Processing an XFDF document into PL/SQL table
function parse_xfdf(
  p_file clob
  ) return t_xml_inputs;

end plpdf_xfdf;
/

CREATE OR REPLACE PACKAGE BODY plpdf_xfdf is
--v2.0.0
procedure start_ 
  is
begin
  plx_sc.startdocument(g_xml_doc);
  plx_sc.startdocumentwheader(g_xml_doc,
                              c_xml_version,
                              c_xml_encoding,
                              c_xml_standalone);
end;
--------------------------------------------------------------------------------------------------------
procedure end_ 
  is
begin
  plx_sc.enddocument(g_xml_doc);
end;
--------------------------------------------------------------------------------------------------------
procedure startxfdf 
  is
  l_attr plx_sc.type_attrs;
begin
  plx_sc.addattribute(l_attr, t_xmlns, c_xmlns);
  plx_sc.addattributewns(l_attr, t_xml, t_space, c_xml_space);
  plx_sc.startelement(g_xml_doc, t_xfdf, l_attr);
end;
--------------------------------------------------------------------------------------------------------
procedure endxfdf 
  is
begin
  plx_sc.endelement(g_xml_doc, t_xfdf);
end;
--------------------------------------------------------------------------------------------------------
procedure ids(
  p_original_id plx_sc.maxv2, 
  p_modified_id plx_sc.maxv2
  ) is
  l_attr plx_sc.type_attrs;
begin
  plx_sc.addattribute(l_attr, t_original, p_original_id);
  plx_sc.addattribute(l_attr, t_modified, p_modified_id);
  plx_sc.startelement(g_xml_doc, t_ids, l_attr, true);
end;
--------------------------------------------------------------------------------------------------------
procedure startfields 
  is
begin
  plx_sc.startelement(g_xml_doc, t_fields);
end;
--------------------------------------------------------------------------------------------------------
procedure endfields 
  is
begin
  plx_sc.endelement(g_xml_doc, t_fields);
end;
--------------------------------------------------------------------------------------------------------
procedure startfield(
  p_name plx_sc.maxv2
  ) is
  l_attr plx_sc.type_attrs;
begin
  plx_sc.addattribute(l_attr, t_name, p_name);
  plx_sc.startelement(g_xml_doc, t_field, l_attr);
end;
--------------------------------------------------------------------------------------------------------
procedure endfield 
  is
begin
  plx_sc.endelement(g_xml_doc, t_field);
end;
--------------------------------------------------------------------------------------------------------
procedure startvalue 
  is
begin
  plx_sc.startelement(g_xml_doc, t_value);
end;
--------------------------------------------------------------------------------------------------------
procedure endvalue 
  is
begin
  plx_sc.endelement(g_xml_doc, t_value);
end;
--------------------------------------------------------------------------------------------------------
procedure value(
  p_value plx_sc.maxv2
  ) is
begin
  plx_sc.addtext(g_xml_doc, p_value);
end;
--------------------------------------------------------------------------------------------------------
function getblob(
  p_xml_doc clob
  ) return blob as
  l_xml_doc  clob := p_xml_doc;
  l_xml_blob blob;
begin
  plx_sc.getblob(l_xml_doc, l_xml_blob);
  return l_xml_blob;
end;
--------------------------------------------------------------------------------------------------------
function create_xfdf(
  p_xml_inputs t_xml_inputs,
  p_id         varchar2 default null
  ) return blob as
  l_xml_blob blob;
  l_name   plx_sc.normalv2;
begin
  if p_xml_inputs.count > 0 then    
    start_;   
    startxfdf;   
    ids(p_id, p_id);
    startfields;   
    l_name := p_xml_inputs.first;    
    while (l_name is not null) loop     
      startfield(l_name);     
      startvalue;
      value(p_xml_inputs(l_name));
      endvalue;     
      endfield;     
      l_name := p_xml_inputs.next(l_name);      
    end loop;   
    endfields;   
    endxfdf;   
    end_; 
    l_xml_blob := getblob(g_xml_doc);  
  end if;
  --
  return l_xml_blob;
end;
--------------------------------------------------------------------------------------------------------
function parse_xfdf(
  p_file clob
  ) return t_xml_inputs is
  l_xml_ret        t_xml_inputs;
  l_xml_doc        xmldom.domdocument;
  l_node_list    xmldom.domnodelist;
  l_len1         number;
  l_len2         number;
  l_node         xmldom.domnode;
  l_element      xmldom.domelement;
  l_element_name plx_sc.normalv2;
  l_node_map     xmldom.domnamednodemap;
  l_attrval      plx_sc.normalv2;
  
begin
  l_xml_doc := dbms_xmldom.newdomdocument(p_file); 
  -- get all elements
  l_node_list := xmldom.getelementsbytagname(l_xml_doc, '*'); 
  l_len1 := xmldom.getlength(l_node_list); 
  -- loop through elements
  for j in 0 .. l_len1 - 1 loop   
    l_node         := xmldom.item(l_node_list, j);
    l_element      := xmldom.makeelement(l_node);
    l_element_name := xmldom.gettagname(l_element);
    --  
    if l_element_name in (t_field, t_value) then
      -- get all attributes of element
      l_node_map := xmldom.getattributes(l_node);     
      if not(xmldom.isnull(l_node_map)) then
        l_len2 := xmldom.getlength(l_node_map);       
        -- loop through attributes
        for i in 0 .. l_len2 - 1 loop
          l_node := xmldom.item(l_node_map, i);         
          if xmldom.getnodename(l_node) = t_name then
            l_attrval := xmldom.getnodevalue(l_node);
          end if;         
        end loop;       
      end if;     
      l_node := xmldom.getfirstchild(l_node);     
      if xmldom.getnodetype(l_node) = xmldom.text_node then
        l_xml_ret(l_attrval) := xmldom.getnodevalue(l_node);
      end if;
    end if;  
  end loop;
  --
  return l_xml_ret;
end;
--
end plpdf_xfdf;
/
